//
// Class:
//  component_nasd
//
//  Description:
//   [see corresponding header file]
//
//

// $Id: nasdsim_component_nasd.C,v 1.24 2000/05/10 19:15:09 bforney Exp $

#include <math.h>
#include <float.h> /**/
#include <algorithm>
#include "nasdsim_component_nasd.h"
#include "nasdsim_debug.h"

//*****
//
// component_nasd
//

//
// component_nasd::component_nasd() [constructor]
//
// description:
//  constructor a new component_nasd instance. Configure with disk
//  bandwidth, average seek time, cache size, and block size.
//
// parameters:
//  component_id - identifier for this object (only needs to be unique among
//                 NASDs this should be the same as the NASD's "network id"
//  disk_bandwidths - bandwidth in sim_bandwidth_t units
//  avg_seek_time - time for the disk to seek to a position on average
//  cache_size - size of the cache in bytes
//  block_size - size of a disk/cache block in bytes
//
component_nasd::component_nasd(
			       sim_component_id_t component_id,
			       disk_bandwidths_t *disk_bandwidths,
			       simtime_t avg_seek_time,
			       size_t cache_size,
			       size_t block_size,
			       sim_bandwidth_t link_bandwidth,
			       sim_len_t link_header_size,
			       perceived_map_t &perceived_bw,
			       int rotational_speed
			       )
  : disk_bandwidths_(disk_bandwidths), rot_speed_(rotational_speed),
    perceived_bw_(perceived_bw)
{
  component_id_ = component_id;
  avg_seek_time_ = avg_seek_time;
  next_request_begin_time_ = 0;
  block_size_ = block_size;
  stats_total_requests_ = 0;
  prev_object_id_ = 0;
  prev_offset_ = 0;

  link = new component_link (component_id_, 
			     link_bandwidth,
			     link_header_size);
			     
  cur_disk_bandwidth_ = disk_bandwidths_->front().bandwidth;
  disk_bandwidths->pop();
  cache_size_ = cache_size;
  
#if defined(NASD_LOCAL_CACHE)
 cache_local_ = new component_nasd_cache_local(block_size);
 if (cache_local_ == NULL) {
   throw;
 }
#if defined(NASD_REMOTE_CACHING)
 perceived_bw_[component_id_].perceived_bw = LDBL_MAX;
 perceived_bw_[component_id_].nasd_id = component_id_;
 cache_global_ = new component_nasd_cache_global(block_size);
 if (cache_global_ == NULL) {
   throw;
 }
 stats_remote_hits_ = 0;
 stats_remote_misses_ = 0;
 stats_remote_returned_ = 0;
#endif
#endif
}

//
// component_nasd::~component_nasd() [destructor]
//
// description:
//  destructor to clean up any dynamically managed memory for instances
//  of this object
//
// parameters:
//  none
//
component_nasd::~component_nasd()
{
#if defined(NASD_LOCAL_CACHE)
  if (cache_local_ != NULL) {
    delete cache_local_;
  }
#if defined(NASD_REMOTE_CACHING)
  if (cache_global_ != NULL) {
    delete cache_global_;
  }
#endif
#endif
  delete link;
}

simtime_t
component_nasd::disk_io(
			block_list_t *block_list,
			bool write,
			simtime_t start_time
			)
{
  sim_offset_t prev_subrequest;
  simtime_t completion_time = 0;

  // read from disk any subrequests (request for individual blocks of an
  // object) which missed in the cache. Try to be realistic about layout.
  if (prev_object_id_ != block_list->front().object_id_ ||
      prev_offset_ +  block_size_ != block_list->front().block_id_) {
    completion_time += avg_seek_time_; // add the first seek. the disk head
                                       // could be anywhere, so assume average
                                       // seek time.
  }
  completion_time += (0.5 * 60.0)/rot_speed_; // add average rotational delay
  
  prev_subrequest = 0;

  for (block_list_t::iterator iter = block_list->begin();
       iter != block_list->end(); iter++) {
    // check if there is a discontinuity in the subrequest. If there is,
    // assume a small seek occurs since object blocks should be laid out
    // fairly close to each other.
    if (iter->block_id_ - prev_subrequest > block_size_ &&
	prev_subrequest != 0) {
      completion_time += avg_seek_time_/3 + (0.5 * 60.0)/rot_speed_;
    }

    // update the disk bandwidth if needed
    if (!disk_bandwidths_->empty() &&
	(disk_bandwidths_->front().time <= start_time + completion_time)) {
      cur_disk_bandwidth_ = disk_bandwidths_->front().bandwidth;
#if defined(DEBUG)
      cout << "CHANGED DISK BANDWIDTH TO " << cur_disk_bandwidth_ << endl;
#endif
      disk_bandwidths_->pop();
    } 
    completion_time += block_size_/cur_disk_bandwidth_;
      
    // set up for the next iteration
    prev_subrequest = iter->block_id_;

    prev_object_id_ = iter->object_id_;
    prev_offset_ = iter->block_id_;
  }

  return completion_time;
}

//
// component_nasd::dump_stats()
//
// description:
//  dump any statistics for a nasd component object
//
// parameters:
//  output - the ostream to which the stats should be dumped
//
// return values:
//  none
//
void
component_nasd::dump_stats(
			   ostream &output
			   ) const
{
  output << "NASD #" << component_id_ << endl;
  output << "total requests: " << stats_total_requests_ << endl;
  output << "perceived bandwidth: ";
  if (perceived_bw_[component_id_].xfer_time != 0) {
    output << (perceived_bw_[component_id_].bytes_xfered /
	       perceived_bw_[component_id_].xfer_time);
  } else {
    output << 0;
  }
  output << " bytes / second" << endl;

  output << "cache: (at the block granularity)" << endl;
  cache_local_->dump_stats(output);
#if defined(NASD_REMOTE_CACHING)
  output << "remote (global cache):" << endl;
  output << "   hits:                  " << stats_remote_hits_ << endl;
  output << "   misses:                " << stats_remote_misses_ << endl;
  output << "   returned (false hits): " << stats_remote_returned_ << endl;
  output << "   adds:                  " << stats_remote_adds_ << endl;
  output << "   removes:               " << stats_remote_removes_ << endl;
  cache_global_->dump_stats(output);
#endif
  output << endl;
}

component_link *
component_nasd::getLink (void) 
{
  return link;
}

//
// component_nasd::compute_perceived_bw()
//
// description:
//  Compute the perceived bandwidth of the NASD. After computing the perceived
//  bandwidth, reset the bytes_xfered and xfer_time to accumulate new values
//  for the next call to this method. Use a smoothing function to keep large
//  variations in values from time to time.
//
// parameters:
//  none
//
// return values:
//  none
//
void
component_nasd::compute_perceived_bw(void)
{
  sim_bandwidth_t cur_bandwidth;

  // if idle during this computation period
  if (perceived_bw_[component_id_].xfer_time == 0) {
    // if we have seen requests before, double our bandwidth to bias
    // other NASDs to use our cache for global caching
    if (perceived_bw_[component_id_].perceived_bw != LDBL_MAX) { 
      perceived_bw_[component_id_].perceived_bw *= 2;
    } else {
      cur_bandwidth = LDBL_MAX;
    }
  } else { // not idle
    cur_bandwidth = perceived_bw_[component_id_].bytes_xfered /
      perceived_bw_[component_id_].xfer_time;
    if (perceived_bw_[component_id_].perceived_bw != LDBL_MAX) {
      perceived_bw_[component_id_].perceived_bw =
	0.8 * cur_bandwidth + 0.2 * perceived_bw_[component_id_].perceived_bw;
    } else {
      perceived_bw_[component_id_].perceived_bw = cur_bandwidth;
    }
    cur_bandwidth = LDBL_MAX;
  }

  // reset
  perceived_bw_[component_id_].bytes_xfered = 0;
   perceived_bw_[component_id_].xfer_time = 0;
}

void
component_nasd::inc_request_count(void)
{
  stats_total_requests_++;
}

void
component_nasd::add_perceived_bw(
				 simtime_t xfer_time,
				 size_t bytes_xfered
				 )
{
  perceived_bw_[component_id_].xfer_time += xfer_time;
  perceived_bw_[component_id_].bytes_xfered += bytes_xfered;
}

void
component_nasd::set_next_request_time(
				      simtime_t new_time
				      )
{
  ASSERT(next_request_begin_time_ <= new_time);
  next_request_begin_time_ = new_time;
}

#if defined(NASD_LOCAL_CACHE) || defined(NASD_REMOTE_CACHING)
bool
component_nasd::is_cache_full(void)
{
  size_t size;

  size = cache_local_->get_size();
#if defined(NASD_REMOTE_CACHING)
  size += cache_global_->get_size();
#endif // defined(NASD_REMOTE_CACHING)

  ASSERT(size <= cache_size_);

  return size == cache_size_;
}
#endif // defined(NASD_LOCAL_CACHE) || defined(NASD_REMOTE_CACHING)


#if defined(NASD_LOCAL_CACHE)

block_list_t *
component_nasd::cache_lookup_local(
				   sim_object_t object,
				   sim_offset_t offset,
				   sim_len_t length,
				   bool write,
				   simtime_t &completion_time
				   )
{
   // update the NASD's time if it lags behind the simulation time
  if (completion_time > next_request_begin_time_) {
    next_request_begin_time_ = completion_time;
  } else {
    completion_time = next_request_begin_time_;
  }
 
  return cache_local_->lookup(object, offset, length, write, completion_time);
}

simtime_t
component_nasd::cache_add_local(
				block_list_t *block_list,
				bool write
				)
{
  return cache_local_->add(block_list, write);

#if defined(NASD_GLOBAL_CACHING)
  ASSERT(cache_local_->get_size() + cache_global_->get_size()
	 <= get_max_cache_size());
#else
  ASSERT(cache_local_->get_size() <= get_max_cache_size());
#endif
}

size_t
component_nasd::cache_evict_local(
				  size_t num_blocks,
				  sim_object_t object,
				  sim_offset_t offset,
				  sim_len_t length,
				  block_list_t &evicted_clean,
				  block_list_t &evicted_dirty
				  )
{
  return cache_local_->evict(num_blocks,
			     object, offset, length,
			     evicted_clean, evicted_dirty);
}

size_t
component_nasd::get_cache_local_size(void)
{
  return cache_local_->get_size();
}

#if defined(NASD_REMOTE_CACHING)
nasd_global_map_t *
component_nasd::cache_lookup_remote(
				    block_list_t *block_list,
				    bool write,
				    simtime_t &completion_time
				    )
{
  nasd_global_map_t *global_map;
  remote_cachers_iter_t rc_iter;
  block_cachers_iter_t bc_iter;
  block_info_t block_info;
  sim_object_t object_id = block_list->front().object_id_;

  rc_iter = remote_cachers_tbl_.find(object_id);
  if (rc_iter != remote_cachers_tbl_.end()) {
    global_map = new nasd_global_map_t;    
    for (block_list_t::iterator bl_iter = block_list->begin(), bl_iter_old;
	 bl_iter != block_list->end();) {
      bc_iter = rc_iter->second.find(bl_iter->block_id_);
      // if the entry was found in a remote cache, do two things:
      // 1. add it to the global_map so that event(s) can be generated
      //    to get the entry
      // 2. remove the entry from the block_list (prune)
      if (bc_iter != rc_iter->second.end()) {
	stats_remote_hits_++;

	block_info.object_id_ = object_id;
	block_info.block_id_ = bl_iter->block_id_;
	(*global_map)[bc_iter->second].push_back(block_info);

	bl_iter_old = bl_iter++;
	block_list->erase(bl_iter_old);

#if defined(DEBUG)
	cout << "component_nasd::cache_lookup_remote() hit obj = "
	     << block_info.object_id_ << ", block_id = "
	     << block_info.block_id_ << endl;
#endif
      } else {
	stats_remote_misses_++;

#if defined(DEBUG)
	cout << "component_nasd::cache_lookup_remote() miss obj = "
	     << bl_iter->object_id_ << ", block_id = "
	     << bl_iter->block_id_ << endl;
#endif
	bl_iter++;
      }
    }

    if (global_map->size() == 0) {
      delete global_map;
      global_map = NULL;
    }
  } else {
    global_map = NULL;
  }

  return global_map;
}

void
component_nasd::cache_add_remote(
				 block_list_t *block_list,
				 sim_component_id_t remote_nasd_id
				 )
{
  block_list_t::iterator bl_iter;

  // assume all blocks in the block list are for the same
  // object

  // go through all of the blocks in the block list and add them
  // to the remote cachers table
  for (bl_iter = block_list->begin(); bl_iter != block_list->end();
       bl_iter++) {
    stats_remote_adds_++;
#if defined(DEBUG)
    cout << "component_nasd::cache_add_remote() adding obj = "
	 << bl_iter->object_id_ << ", block_id = " << bl_iter->block_id_
	 << endl;
#endif
    remote_cachers_tbl_[bl_iter->object_id_][bl_iter->block_id_] =
      remote_nasd_id;
  }
}

void
component_nasd::cache_remove_remote(
				    block_list_t &block_list
				    )
{
  block_list_t::iterator bl_iter;
  remote_cachers_iter_t rc_iter;
  sim_object_t object_id = block_list.front().object_id_;

  rc_iter = remote_cachers_tbl_.find(object_id);
  if (rc_iter != remote_cachers_tbl_.end()) {
    for (bl_iter = block_list.begin();
	 bl_iter != block_list.end(); bl_iter++) {
      stats_remote_removes_++;
#if defined(DEBUG)
      cout << "component_nasd::cache_remove_remote(): removing obj = "
	   << object_id << ", block = " << bl_iter->block_id_ << endl;
#endif
      rc_iter->second.erase(bl_iter->block_id_);
    }

    if (rc_iter->second.empty()) {
      remote_cachers_tbl_.erase(rc_iter);
    }
  }
}

void
component_nasd::cache_remove_remote(
				    sim_object_t object,
				    sim_offset_t offset,
				    sim_len_t length
				    )
{
  remote_cachers_iter_t rc_iter;
  sim_offset_t block;

  rc_iter = remote_cachers_tbl_.find(object);
  if (rc_iter != remote_cachers_tbl_.end()) {
    for (block = offset; block < offset + length; block += block_size_) {
      stats_remote_removes_++;
#if defined(DEBUG)
      cout << "component_nasd::cache_remove_remote(): removing obj = "
	   << object << ", block = " << block << endl;
#endif
      rc_iter->second.erase(block);
    }

    if (rc_iter->second.empty()) {
      remote_cachers_tbl_.erase(rc_iter);
    }
  }
}

size_t
component_nasd::get_cache_global_num_entries(sim_component_id_t id)
{
  return cache_global_->num_entires(id);
}

void
component_nasd::cache_add_global(
				 block_list_t *block_list,
				 sim_component_id_t owner_id
				 )
{
  ASSERT(cache_local_->get_size() +
	 cache_global_->get_size() + block_list->size() * block_size_
	 <= get_max_cache_size());  

  cache_global_->add(block_list, owner_id);
  
  ASSERT(cache_local_->get_size() + cache_global_->get_size()
	 <= get_max_cache_size());
}

block_list_t *
component_nasd::cache_lookup_global(
				    block_list_t *block_list,
				    sim_component_id_t owner_id,
				    simtime_t &completion_time
				    )
{
  return cache_global_->lookup(block_list, owner_id, completion_time);
}

void
component_nasd::cache_invalidate_global(
					block_list_t *block_list,
					sim_component_id_t owner_id
					)
{
  cache_global_->invalidate(block_list, owner_id);
}

int
component_nasd::cache_evict_global(
				   int blocks,
				   sim_component_id_t remote_nasd_id
				   )
{
  return cache_global_->evict(blocks, remote_nasd_id);
}

size_t
component_nasd::get_cache_size(void)
{
  return cache_local_->get_size() + cache_global_->get_size();
}
#endif // defined(NASD_REMOTE_CACHING)


// *******
// component_nasd_cache_local

component_nasd_cache_local::component_nasd_cache_local(
						       size_t block_size
						       )
  : block_size_(block_size)
{
  double block_bits;

  block_bits = log10(block_size)/log10(2);
  // make sure the block size is a power of two
  if (block_bits != (long)block_bits) {
    throw;
  }

  offset_tag_mask_ = 0;
  offset_tag_mask_ = ~offset_tag_mask_ << (long)block_bits;

  stats_hits_ = 0;
  stats_misses_ = 0;
  stats_adds_ = 0;
  stats_evicts_ = 0;
}

block_list_t *
component_nasd_cache_local::lookup(
				   sim_object_t object,
				   sim_offset_t offset,
				   sim_len_t length,
				   bool write,
				   simtime_t &completion_time
				   )
{
  block_list_t *unsat_block_list; // blocks that the local cache could not
                                  // satisfy
  object_map_iter_t obj_map_iter;
  offset_map_iter_t off_map_iter;
  const double single_lookup_time = 100*0.001*0.001*0.001;
  sim_offset_t cur_position;
  block_info_t entry;
  size_t cache_size;
  cache_entry_t c_entry;

  unsat_block_list = new block_list_t;

  // look up each block
  for (cur_position = (offset/block_size_)*block_size_; // make the request
                                                        // begin on a block
                                                        // boundary
       cur_position < (offset + length);
       cur_position += block_size_) {
    // determine if the block is in the cache
    if (((obj_map_iter = object_map_.find(object)) != object_map_.end()) &&
	((off_map_iter =
	  obj_map_iter->second.find(cur_position & offset_tag_mask_)) !=
	 obj_map_iter->second.end())) {
      stats_hits_++;
#if defined(DEBUG)
      cout << "component_nasd_cache_local::lookup(): cache hit for local object = ";
      cout << object << ", offset = " << cur_position << endl;
#endif
    
      // update LRU list
      c_entry = *(off_map_iter->second.element);
      off_map_iter->second.list->erase(off_map_iter->second.element);
      if (write) {
	// add this to the front of the dirty LRU list
	entries_dirty_.push_front(c_entry);
      } else {
	entries_clean_.push_front(c_entry);
      }
    } else {
      stats_misses_++;
#if defined(DEBUG)
      cout << "component_nasd_cache_local::lookup(): cache miss for local object = ";
      cout << object << ", offset = " << cur_position << endl;
#endif

      // add this cache miss to the unsat_block_list
      entry.object_id_ = object;
      entry.block_id_ = cur_position;
      unsat_block_list->push_back(entry);
    }

    cache_size = entries_clean_.size() + entries_dirty_.size();
    if (cache_size <= 2) {
      completion_time += single_lookup_time;
    } else {
      completion_time += (ceil(log(cache_size)/log(2)))*single_lookup_time;
    }
  }

  if (unsat_block_list->size() == 0) {
    delete unsat_block_list;
    unsat_block_list = NULL;
  }

  return unsat_block_list;
}

simtime_t
component_nasd_cache_local::add(
				block_list_t *block_list,
				bool write
				)
{
  cache_entry_t entry;

  for (block_list_t::iterator iter = block_list->begin();
       iter != block_list->end(); iter++) {
    stats_adds_++;
    entry.object = iter->object_id_;
    entry.offset_tag = iter->block_id_ & offset_tag_mask_;
#if defined(DEBUG)
    cout << "component_nasd_cache_local::add(): adding object = ";
    cout << entry.object << ", offset = " << entry.offset_tag << endl;
#endif
    if (write) {
      entries_dirty_.push_front(entry);
      object_map_[entry.object][entry.offset_tag].list = &entries_dirty_;
      object_map_[entry.object][entry.offset_tag].element =
	entries_dirty_.begin();
    } else {
      entries_clean_.push_front(entry);
      object_map_[entry.object][entry.offset_tag].list = &entries_clean_;
      object_map_[entry.object][entry.offset_tag].element =
	entries_clean_.begin();
    }
  }

  return 0;
}


//
// This isn't true LRU since blocks that are in the request and the cache
// are moved around in the LRU list. However, this is pretty close to true
// LRU.
//
size_t
component_nasd_cache_local::evict(
				  size_t num_blocks,
				  sim_object_t object,
				  sim_offset_t offset,
				  sim_len_t length,
				  block_list_t &evicted_clean,
				  block_list_t &evicted_dirty
				  )
{
  cache_entry_t entry;
  block_info_t block_info;
  object_map_iter_t obj_map_iter;
  size_t num_blocks_initial = num_blocks;

  while (num_blocks > 0
#if defined(NASD_REMOTE_CACHING)
	 && (!entries_clean_.empty() ||
	 !entries_dirty_.empty())
#endif // defined(NASD_REMOTE_CACHING)
	 ) {
    if (!entries_clean_.empty()) {
      entry = entries_clean_.back();
#if defined(NASD_GLOBAL_CACHING)
      // ensure that we don't evict blocks that are part of the current
      // request (this is imiportant for global caching)
      if (entry.object == object &&
	  entry.offset_tag >= offset && entry.offset_tag < offset + length) {
	entries_clean_.push_front(entries_clean_.back());
	entries_clean_.pop_back();
      } else {
#endif // defined(NASD_REMOTE_CACHING)
	entries_clean_.pop_back();

	block_info.object_id_ = entry.object;
	block_info.block_id_ = entry.offset_tag;
	evicted_clean.push_back(block_info);
	stats_evicts_++;

	num_blocks--;
#if defined(NASD_GLOBAL_CACHING)
      }
#endif // defined(NASD_REMOTE_CACHING)
    } else {
      entry = entries_dirty_.back();
#if defined(NASD_GLOBAL_CACHING)
      // ensure that we don't evict blocks that are part of the current
      // request
      if (entry.object == object &&
	  entry.offset_tag >= offset && entry.offset_tag < offset + length) {
	entries_dirty_.push_front(entries_dirty_.back());
	entries_dirty_.pop_back();
      } else {
#endif // defined(NASD_REMOTE_CACHING)
	entries_dirty_.pop_back();

	block_info.object_id_ = entry.object;
	block_info.block_id_ = entry.offset_tag;
	evicted_dirty.push_back(block_info);
	stats_evicts_++;

	num_blocks--;
#if defined(NASD_GLOBAL_CACHING)
      }
#endif // defined(NASD_REMOTE_CACHING)
    }

    // remove from the lookup map, if the item was really removed
    if (!(entry.object == object &&
	  entry.offset_tag >= offset &&
	  entry.offset_tag < offset + length)) {
      obj_map_iter = object_map_.find(entry.object);
      obj_map_iter->second.erase(obj_map_iter->second.find(entry.offset_tag));
      if (obj_map_iter->second.empty()) {
	object_map_.erase(obj_map_iter);
      }
#if defined(DEBUG)
      cout << "evicted local: object = " << entry.object << ", tag = " <<
	entry.offset_tag << endl;
#endif
    }
  }

  ASSERT(num_blocks_initial - num_blocks ==
	 evicted_dirty.size() + evicted_clean.size());

  return num_blocks_initial - num_blocks;
}

void
component_nasd_cache_local::dump_stats(
				       ostream &output
				       )
{
  output << "local cache:" << endl;
  output << "   hits:   " << stats_hits_ << endl;
  output << "   misses: " << stats_misses_ << endl;
  output << "   adds:   " << stats_adds_ << endl;
  output << "   evicts: " << stats_evicts_ << endl;
}

// *****
// component_nasd_cache_global

#if defined(NASD_REMOTE_CACHING)
component_nasd_cache_global::component_nasd_cache_global(
							 size_t block_size
							 )
  : block_size_(block_size)
{
  double block_bits;

  block_bits = log10(block_size)/log10(2);
  // make sure the block size is a power of two
  if (block_bits != (long)block_bits) {
    throw;
  }

  offset_tag_mask_ = 0;
  offset_tag_mask_ = ~offset_tag_mask_ << (long)block_bits;

  stats_hits_ = 0;
  stats_misses_ = 0;
  stats_invalidations_ = 0;
  stats_adds_ = 0;
  stats_evicts_ = 0;
}

//
// returns a list of satisfied block request for a "remote/forwarded"
// read.
//
// block_list will contain all unsatisfied requests
//
//
block_list_t *
component_nasd_cache_global::lookup(
				    block_list_t *block_list,
				    sim_component_id_t owner_nasd_id,
				    simtime_t &completion_time
				    )
{
  block_list_t *sat_block_list = NULL;
  block_list_t::iterator bl_iter, bl_iter_old;
  table_iter_t tbl_iter;
  cache_list_iter_t cache_l_iter;
  cache_entry_t entry;
 
  tbl_iter = table_.find(owner_nasd_id);
  if (tbl_iter != table_.end()) {
    sat_block_list = new block_list_t;

    for(bl_iter = block_list->begin(); bl_iter != block_list->end();) {
      entry.object = bl_iter->object_id_;
      entry.offset_tag = bl_iter->block_id_;

      // if the entry is found in the global cache, remove from the block
      // list and add to the satisfied list.
      for (cache_l_iter = tbl_iter->second.begin();
	   cache_l_iter != tbl_iter->second.end() &&
	     cache_l_iter->object != bl_iter->object_id_ &&
	     cache_l_iter->offset_tag != bl_iter->block_id_;
	   cache_l_iter++);
      if (cache_l_iter != tbl_iter->second.end()) {
	stats_hits_++;
#if defined(DEBUG)
	cout << "component_nasd_cache_global::lookup(): cache hit for local object = ";
	cout << bl_iter->object_id_ << ", offset = " << bl_iter->block_id_ << endl;
#endif
	// add to the satisfied list and remove from block_list, which is
	// the unsatisfied list
	sat_block_list->push_back(*bl_iter);
	bl_iter_old = bl_iter++;
	block_list->erase(bl_iter_old);

	// update the LRU information by modifying the cache list
	entry = *cache_l_iter;
	tbl_iter->second.erase(cache_l_iter);
	tbl_iter->second.push_front(entry);
      } else {
	stats_misses_++;
#if defined(DEBUG)
	cout << "component_nasd_cache_global::lookup(): cache miss for local object = ";
	cout << bl_iter->object_id_ << ", offset = " << bl_iter->block_id_ << endl;
#endif
	bl_iter++;
      }
    }
  }

  completion_time += 0.001*0.001;

  return sat_block_list;
}

void
component_nasd_cache_global::add(
				 block_list_t *block_list,
				 sim_component_id_t owner_nasd_id
				 )
{
  block_list_t::iterator bl_iter;
  cache_entry_t entry;

  for (bl_iter = block_list->begin();
       bl_iter != block_list->end(); bl_iter++) {
    stats_adds_++;
    entry.object = bl_iter->object_id_;
    entry.offset_tag = bl_iter->block_id_;
#if defined(DEBUG)
    cout << "component_nasd_cache_global::add(): object = ";
    cout << bl_iter->object_id_ << ", offset = " << bl_iter->block_id_ << endl;
#endif

    table_[owner_nasd_id].push_front(entry);
  }
}

void
component_nasd_cache_global::invalidate(
					block_list_t *block_list,
					sim_component_id_t owner_nasd_id
					)
{
  block_list_t::iterator bl_iter;
  cache_entry_t entry;
  cache_list_t cl;
  cache_list_iter_t cl_iter;

  ASSERT(table_.find(owner_nasd_id) != table_.end());
  cl = table_[owner_nasd_id];

  for (bl_iter = block_list->begin(); bl_iter != block_list->end();
       bl_iter++) {
    stats_invalidations_++;
    entry.object = bl_iter->object_id_;
    entry.offset_tag = bl_iter->block_id_;

    cl_iter = find(cl.begin(), cl.end(), entry);

    if (cl_iter != cl.end()) {
      cl.erase(cl_iter);
    }
  }
}

// return the number of blocks left, so blocks - number of evicted blocks
int
component_nasd_cache_global::evict(
				   int blocks,
				   sim_component_id_t remote_nasd_id
				   )
{
  table_iter_t iter;

  iter = table_.find(remote_nasd_id);

  // if this nasd has any blocks, evict
  if (iter != table_.end()) {
    while (blocks > 0 && !iter->second.empty()) {
      stats_evicts_++;
#if defined(DEBUG)
      cout << "component_nasd_cache_global::evict() object = "
	   << iter->second.back().object << ", offset = "
	   << iter->second.back().offset_tag << endl;
#endif
      iter->second.pop_back();
      blocks--;
    }
  }

  return blocks;
}

// returns in block size
int
component_nasd_cache_global::num_entires(
					 sim_component_id_t nasd_id
					 )
{
  table_iter_t iter;

  if ((iter = table_.find(nasd_id)) != table_.end()) {
    return iter->second.size() * block_size_;
  } else {
    return 0;
  }
}

size_t
component_nasd_cache_global::get_size(void)
{
  table_iter_t iter;
  size_t size = 0;

  for (iter = table_.begin(); iter != table_.end(); iter++) {
    size += iter->second.size();
  }

  return size * block_size_;
}

void
component_nasd_cache_global::dump_stats(
					ostream &output
					)
{
  output << "global cache (@ this NASD):" << endl;
  output << "   hits:          " << stats_hits_ << endl;
  output << "   misses:        " << stats_misses_ << endl;
  output << "   invalidations: " << stats_invalidations_ << endl;
  output << "   adds:   " << stats_adds_ << endl;
  output << "   evicts: " << stats_evicts_ << endl;
}

#endif // defined(NASD_REMOTE_CACHING)
#endif // defined(NASD_LOCAL_CACHE)


