// $Id: nasdsim_config.C,v 1.6 2000/05/05 21:20:04 ozaki Exp $

#include <strstream.h>
#include <iomanip.h>
#include <strings.h>
#include <iostream>
#include "nasdsim_config.h"

#define COMMENT_SZ 100
#define MAX_LINE_SZ 256

nasdsim_config::nasdsim_config (char *filename)
{
  nwInfo = NULL;
  nasdList = NULL;
  clientList = NULL;
  
  ifstream fin (filename);
  
  if (!fin) {
    throw;
  }  

  parseFile (fin);
}

nasdsim_config::~nasdsim_config (void)
{
  if (nwInfo != NULL) delete nwInfo;
  
  if (nasdList != NULL) cleanUp (nasdList);
  
  if (clientList != NULL) cleanUp (clientList);
}

void 
nasdsim_config::print (nasdInfo *nInfo)
{
  cout << "nasdInfo {" 
       << "id = " << nInfo->id << ", "
       << "seekTime = " << nInfo->seekTime << ", "
       << "cacheSize = " << nInfo->cacheSize << ", "
       << "blockSize = " << nInfo->blockSize << ", "
       << "queueSize = " << nInfo->queueSize << "}" 
       << endl;
}

void 
nasdsim_config::print (networkInfo *nwInfo)
{
  cout << "networkInfo {" 
       << "bw = " << nwInfo->bandwidth << ", "
       << "headerSize = " << nwInfo->headerSize << ", "
       << "switchTime = " << nwInfo->switchTime << "}"
       << endl;
}

void 
nasdsim_config::print (clientInfo *cInfo)
{
  cout << "clientInfo {" 
       << "id = " << cInfo->id << ", "
       << "filename = " << cInfo->filename << "}"
       << endl;
}

void 
nasdsim_config::parseFile (ifstream &fin) 
{
  char buf [MAX_LINE_SZ + 1];
  bool error = false;
  
  while (fin.getline (buf, MAX_LINE_SZ) && !error) 
    {
      if (strstr (buf, "General") == buf) {
	getGeneralInfo (fin);
      }
      else if (strstr (buf, "Network") == buf) {
	if ((nwInfo = getNetworkInfo (fin)) == NULL) error = true;
      }
      else if (strstr (buf, "Nasds") == buf) {
	if ((nasdList = getNasdList (fin)) == NULL) error = true;
      }
      else if (strstr (buf, "Clients") == buf) {
	if ((clientList = getClientList (fin)) == NULL) error = true;
      }
      else error = true;
    }
}

/**
   static
*/
void *
nasdsim_config::getGeneralInfo (ifstream &fin)
{
  char buf [MAX_LINE_SZ + 1];
  
  if (!fin.getline (buf, MAX_LINE_SZ)) return NULL;

  if (!fin.getline (buf, MAX_LINE_SZ)) return NULL;

  if (!fin.getline (buf, MAX_LINE_SZ)) return NULL;
  
  return NULL;
}


/**
   static
   
   Assumes the following structure:
   
   1 -------
   2 Link_Bandwidth      <bytes/second>
   3 Link_Header_Size    <bytes>
   4 Switch_Time         <seconds>
   5 ##
*/
networkInfo *
nasdsim_config::getNetworkInfo (ifstream &fin)
{
  char buf [MAX_LINE_SZ + 1], comment [COMMENT_SZ + 1];
  
  networkInfo nI, *output;
  
  if (!fin.getline (buf, MAX_LINE_SZ)) return NULL;
  
  // bandwidth
  if (!fin.getline (buf, MAX_LINE_SZ)) return NULL;
  istrstream bwIs (buf, MAX_LINE_SZ);
  bwIs >> setw(COMMENT_SZ) >> comment;
  if (!(bwIs >> nI.bandwidth)) return NULL;
  
  // header size
  if (!fin.getline (buf, MAX_LINE_SZ)) return NULL;
  istrstream hsIs (buf, MAX_LINE_SZ);
  hsIs >> setw (COMMENT_SZ) >> comment;
  if (!(hsIs >> nI.headerSize)) return NULL;
  
  // switch time
  if (!fin.getline (buf, MAX_LINE_SZ)) return NULL;
  istrstream swIs (buf, MAX_LINE_SZ);
  swIs >> setw (COMMENT_SZ) >> comment;
  if (!(swIs >> nI.switchTime)) return NULL;
  
  if (!fin.getline (buf, MAX_LINE_SZ)) return NULL;
  
  output = new networkInfo;
  output->bandwidth = nI.bandwidth;
  output->headerSize = nI.headerSize;
  output->switchTime = nI.switchTime;

  return output;
}

/**
   static
*/
list <nasdInfo *> *
nasdsim_config::getNasdList (ifstream &fin)
{
  char buf [MAX_LINE_SZ + 1], comment [COMMENT_SZ + 1];
  
  if (!fin.getline (buf, MAX_LINE_SZ)) return NULL;
  
  // number of nasds
  int num;
  if (!fin.getline (buf, MAX_LINE_SZ)) return NULL;
  istrstream nIs (buf, MAX_LINE_SZ);
  nIs >> setw (COMMENT_SZ) >> comment;
  if (!(nIs >> num)) return NULL;
  
  if (!fin.getline (buf, MAX_LINE_SZ)) return NULL;
  
  if (num < 1) return NULL;
  
  list <nasdInfo *> *nList = new list <nasdInfo *>;
  

  for (int i = 0; i < num; i++) {
    nasdInfo *nI = getNasdInfo (fin);
    if (nI == NULL) {
      cleanUp (nList); return NULL;
    }
    else
      nList->push_front (nI);     
  }
  
  if (!fin.getline (buf, MAX_LINE_SZ)) return NULL;
  
  return nList;
}

/**
   static
   
   Assumes the following structure:
   
   1 Nasd                <int>
   2 Bandwidth
     time1 bandwidth1
     time2 bandwidth2
     [...]
     timeN bandwidthN
     end
   3 Average_Seek_Time   <seconds>
   4 Cache_Size          <bytes>
   5 Block_Size          <bytes>
   6 Request_Queue_Size  <int>
   7 ----------------------
   
   The order of the bandwidth entries must be in ascending time.
*/
nasdInfo *
nasdsim_config::getNasdInfo (ifstream &fin)
{
  char buf [MAX_LINE_SZ + 1], comment [COMMENT_SZ + 1];
  nasdInfo nI, *output;
  disk_bandwidth_t bw;
  
  // id
  if (!fin.getline (buf, MAX_LINE_SZ)) return NULL;
  istrstream idIs (buf, MAX_LINE_SZ);
  idIs >> setw (COMMENT_SZ) >> comment;
  if (!(idIs >> nI.id)) return NULL;
  
  // bandwidth
  if (!fin.getline (buf, MAX_LINE_SZ)) return NULL;
  istrstream bwIs (buf, MAX_LINE_SZ);
  bwIs >> setw (COMMENT_SZ) >> comment;
  if (!fin.getline(buf, MAX_LINE_SZ)) return NULL;
  nI.bandwidths = new disk_bandwidths_t;
  while (strcasecmp(buf, "end")) {
    istrstream bwtimels(buf, MAX_LINE_SZ);
    bwtimels >> bw.time >> bw.bandwidth;
    if (!bwtimels.fail()) {
      nI.bandwidths->push(bw);
    } else {
      cerr << "error: could not parse bandwidth time correctly for nasd id = " << nI.id << endl;
      delete nI.bandwidths;
      return NULL;
    }
    if (!fin.getline(buf, MAX_LINE_SZ)) {
      delete nI.bandwidths;
      return NULL;
    }
  }
  
  // seek time
  if (!fin.getline (buf, MAX_LINE_SZ))  {
      delete nI.bandwidths;
      return NULL;
  }
  istrstream stIs (buf, MAX_LINE_SZ);
  stIs >> setw (COMMENT_SZ) >> comment;
  if (!(stIs >> nI.seekTime))  {
      delete nI.bandwidths;
      return NULL;
  }
  
  // cache size
  if (!fin.getline (buf, MAX_LINE_SZ))  {
      delete nI.bandwidths;
      return NULL;
  }
  istrstream csIs (buf, MAX_LINE_SZ);
  csIs >> setw (COMMENT_SZ) >> comment;
  if (!(csIs >> nI.cacheSize))  {
      delete nI.bandwidths;
      return NULL;
  }
  
   // block size
  if (!fin.getline (buf, MAX_LINE_SZ)) {
      delete nI.bandwidths;
      return NULL;
  }
  istrstream bsIs (buf, MAX_LINE_SZ);
  bsIs >> setw (COMMENT_SZ) >> comment;
  if (!(bsIs >> nI.blockSize)) {
      delete nI.bandwidths;
      return NULL;
  }
  
  // queue size
  if (!fin.getline (buf, MAX_LINE_SZ)) {
      delete nI.bandwidths;
      return NULL;
  }
  istrstream qsIs (buf, MAX_LINE_SZ);
  qsIs >> setw (COMMENT_SZ) >> comment;
  if (!(qsIs >> nI.queueSize)) {
      delete nI.bandwidths;
      return NULL;
  }
  
  // trailing line
  if (!fin.getline (buf, MAX_LINE_SZ)) {
      delete nI.bandwidths;
      return NULL;
  }

  output = new nasdInfo;

  output->id = nI.id;
  output->bandwidths = nI.bandwidths;
  output->seekTime = nI.seekTime;
  output->cacheSize = nI.cacheSize;
  output->blockSize = nI.blockSize;
  output->queueSize = nI.queueSize;
  
  return output;
}

/**
   static
 */
list <clientInfo *> *
nasdsim_config::getClientList (ifstream &fin)
{
  char buf [MAX_LINE_SZ + 1], comment [COMMENT_SZ + 1];
  
  if (!fin.getline (buf, MAX_LINE_SZ)) return NULL;
  
  // number of clients
  int num;
  if (!fin.getline (buf, MAX_LINE_SZ)) return NULL;
  istrstream nIs (buf, MAX_LINE_SZ);
  nIs >> setw (COMMENT_SZ) >> comment;
  if (!(nIs >> num)) return NULL;
  
  if (!fin.getline (buf, MAX_LINE_SZ)) return NULL;
  
  if (num < 1) return NULL;
  
  list <clientInfo *> *cList = new list <clientInfo *>;
  
  for (int i = 0; i < num; i++) {
    clientInfo *cI = getClientInfo (fin);
    if (cI == NULL) {
      cleanUp (cList); return NULL;
    }
    else
      cList->push_front (cI);     
  }
  
  if (!fin.getline (buf, MAX_LINE_SZ)) return NULL;
  
  return cList;
}

/**
   static
   
   Assumes the following structure:

   1 Client               <int>
   2 Tracefile_Name       <string>
   3 ----------------------
*/
clientInfo *
nasdsim_config::getClientInfo (ifstream &fin)
{
  char buf [MAX_LINE_SZ + 1], comment [COMMENT_SZ + 1];
  
  clientInfo cI, *output;
  
  // id
  if (!fin.getline (buf, MAX_LINE_SZ)) return NULL;
  istrstream idIs (buf, MAX_LINE_SZ);
  idIs >> setw (COMMENT_SZ) >> comment;
  if (!(idIs >> cI.id)) return NULL;
  
  // tracefile
  if (!fin.getline (buf, MAX_LINE_SZ)) return NULL;
  istrstream trIs (buf, MAX_LINE_SZ);
  trIs >> setw (COMMENT_SZ) >> comment;
  if (!(trIs >> setw (CONFIG_TRACE_FILENAME_SZ) >> cI.filename))
    return NULL;
 
  // trailing line
  if (!fin.getline (buf, MAX_LINE_SZ)) return NULL;
  
  output = new clientInfo;
  output->id = cI.id;
  strcpy (output->filename, cI.filename);
  
  return output;
}

/**
   Static
*/
void 
nasdsim_config::cleanUp (list <nasdInfo *> *nList)
{
  list <nasdInfo *>::iterator nListIter;
  
  for (nListIter = nList->begin (); nListIter != nList->end (); nListIter++)
    delete *nListIter;
  
  delete nList;
}

/**
   Static
*/
void 
nasdsim_config::cleanUp (list <clientInfo *> * cList)
{
  list <clientInfo *>::iterator cListIter;
  
  for (cListIter = cList->begin (); cListIter != cList->end (); cListIter++)
    delete *cListIter;
  
  delete cList;
}


