//
// Class:
//  event
//
//  Description:
//   [see associated header file]
//
// Class:
//  event_empty
//
//  Description:
//   [see associated header file]
//

// $Id: nasdsim_event.C,v 1.7 2000/04/24 18:50:57 bforney Exp $

#include "nasdsim_debug.h"
#include "nasdsim_event.h"
#include "nasdsim_param.h"

// *****
// event class
//

//
// event::event() [constructor]
//
// description:
//  constructor to set the starting simulation time
//
// parameters:
//  start_time - starting simulation time for this event
//  inject_time - time the request that this event is for was injected into
//                the simulator. Useful for computing request service time.
//
event::event(
	     simtime_t start_time,
	     simtime_t inject_time
	     )
{
  start_time_ = start_time;
  injection_time_ = inject_time;
}

//
// event::event() [constructor]
//
// description:
//  constructor to set the starting simulation time
//
// parameters:
//  copy_event - an event object to copy
//  start_time - time the event should start
//
event::event(
	     const event *copy_event,
	     simtime_t start_time
	     )
{
  if (copy_event == NULL) {
    injection_time_ = start_time;
  } else {
    injection_time_ = copy_event->injection_time_;
  }
  start_time_ = start_time;
}

//
// event::~event() [destructor]
//
// description:
//  null destructor
//
// parameters:
//  none
//
event::~event()
{
}

//
// event::get_start_time()
//
// description:
//  return the start time of the event
//
// parameters:
//  none
//
// return values:
//  starting time as a positive value
//
inline simtime_t
event::get_start_time(void) const
{
  return start_time_;
}

//
// event::operator<()
//
// description:
//  compares this event against another event to see if this has an earlier
//  start time than the other event. (Used to build the heap.)
//
// parameters:
//  right - the other event to compare against
//
// return values:
//  true - this event has an earlier start time
//  false - this event has a start time greater than or equal to the other
//          event start time.
//
inline bool
event::operator< (
		  const event& right
		  ) const
{
  return start_time_ < right.start_time_;
}


// *****
// event_empty class
//

//
// event_empty::event_empty() [constructor]
//
// description:
//  constructor to set the starting simulation time
//
// parameters:
//  copy_event - base event to copy
//  start_time - time the event should start
//  parameter - the parameters for the request (part of this class to ease
//              destruction of the param object)
//
event_empty::event_empty(
			 const event *copy_event,
			 simtime_t start_time,
			 param *parameter
			 ) :
  event(copy_event, start_time)
{
  parameter_ = parameter;
}

event_empty::event_empty(
			 simtime_t start_time
			 ) :
  event (start_time, start_time)
{
  parameter_ = NULL;
}

//
// event_empty::~event_empty() [destructor]
//
// description:
//  null destructor
//
// parameters:
//  none
//
event_empty::~event_empty()
{
  if (parameter_ != NULL) delete parameter_;
}

//
// event_empty::simulate()
//
// description:
//  Simulate an empty event. This should never happen and the ASSERT()
//  should make this painfully obvious.
//
// parameters:
//  comp - a component object
//
// return values:
//  NULL - could not simulate
//  non-NULL - next event(s)
//
events_t *
event_empty::simulate(
		      components &comp
		      )
{
  return NULL;
}

#if defined(DEBUG)
//
// event_empty::print()
//
// description:
//  When DEBUG is enabled, this method display information about the event.
//
// parameters:
//  output - the ostream for display
//
// return values:
//  none
//
void
event_empty::print(
		   ostream &output
		   ) const
{
  output << "event_empty: { start_time = " << start_time_ << " }";
}
#endif // defined(DEBUG)

