//
// Overview:
//  A request like a read from a client to a NASD is represented in the
//  simulator by a chain of events, ending with an empty_event. Each
//  request carries some basic information about the request through
//  each event comprising the request. This basic information is kept in
//  the event class and copied by each event constructor further down
//  the chain.
//
//  Parameters for each request are also copied to each new event in the
//  request event chain. The parameters are held as a pointer to a param
//  class or subclass - probably a subclass. A pointer is used so that there
//  is not the overhead of a object creation and destruction at each event
//  creation point in the request event chain. Only the empty_event should
//  free the param object.
//
// Class:
//  event
//
//  Description:
//   This is an abstract base class for all events in the simulator. All
//   events should inherit from this class or one of its descendants.
//
// Class:
//  event_empty
//
//  Description:
//   event_empty is the null event. It indicates that a chain of events that
//   compose a tracefile request have completed.
//

// $Id: nasdsim_event.h,v 1.7 2000/04/30 20:43:16 ozaki Exp $

#if !defined(NASDSIM_EVENT_H)
#define NASDSIM_EVENT_H

#include <ostream.h>
#include <queue>
#include "nasdsim_types.h"
#include "nasdsim_components.h"
#include "nasdsim_param.h"

class event;
class components;

typedef queue<event *> events_t;

class event {
 public:
  // constructor and destructor
  event(simtime_t start_time, simtime_t inject_time);
  event(const event *copy_event, simtime_t start_time);
  virtual ~event();

  // accessors
  simtime_t get_start_time(void) const;
  simtime_t get_injection_time(void) const { return injection_time_; }
  void set_start_time (simtime_t stTime) { start_time_ = stTime; }
  void set_injection_time (simtime_t inTime) { injection_time_ = inTime; }
  
  // generators
  virtual events_t *simulate(components &comp) = 0;

  bool operator< (const event& right) const;

#if defined(DEBUG)
  virtual void print(ostream& out) const = 0;
#endif

 protected:
  simtime_t start_time_; // time the event should begin
  simtime_t injection_time_; // time when the first event in the chain of
                             // events that compose a request was added to
                             // to the simulator. Useful in computing service
                             // time
};

class event_empty : public event {
 public:
  // constructor and destructor
  event_empty(const event *copy_event, simtime_t start_time,
	      param *parameter);
  event_empty (simtime_t start_time);
  ~event_empty();

  events_t *simulate(components &comp);

#if defined(DEBUG)
  void print(ostream &output) const;
#endif

 protected:
  param *parameter_;
};

#endif // !define(NASDSIM_EVENT_H)





