//
// Core of the simulator. Contains an event loop and injects events from
// one or more trace files.
//

// $Id: nasdsim_main.C,v 1.20 2000/05/11 04:48:23 bforney Exp $

#include <queue>
#include <limits.h>
#include <float.h>
#include "nasdsim_component_nasd.h"
#include "nasdsim_components.h"
#include "nasdsim_debug.h"
#include "nasdsim_event.h"
#include "nasdsim_types.h"
#include "nasdsim_config.h"

simtime_t NASDSIM_SWITCH_TIME;

//
// A simple comparison class to allow a C++ STL priority queue to be used as
// a heap. Additionally, it assumes the data in the heap are pointers, so the
// pointers need to be dereferenced for comparison.
//
class heap_greater : public binary_function<event *, event *, bool> {
public:
  bool operator()(const event *x, const event *y) const
    { return *x > *y; }
};


typedef priority_queue< event *, vector<event *>, heap_greater > event_heap_t;

//
// prototypes
//
int add_events (event_heap_t &event_heap, component_client &client,
		simtime_t tmp_time, simtime_t current_time); /**/
void run_simulation(components &comp, perceived_map_t &nasd_perceived_bw,
		    const char *config_filename);
void print_stats(const components &comp, simtime_t current_time,
		 int total_requests, simtime_t total_request_time);
#if defined(NASD_REMOTE_CACHING)
ofstream *open_cache_stats(const char *config_filename, components &comp);
void dump_cache_stats(ofstream *output, components &comp,
		      simtime_t current_time);
#endif


int
main(int argc, char *argv[])
{
  components comp; // components of the system that will be simulated
  // hash table of perceived bandwidths of each NASD. This does not take
  // into account all of the network effects on perceived bandwidth
  perceived_map_t nasds_perceived_bw;
  nasdsim_config *config;

  // process command line arguments
  if (argc != 2) {
    exit(1);
  }
  
  // obtain info from config file
  try {
    config = new nasdsim_config (argv [1]);
  } catch (...) {
    cerr << "error: could not open tracefile " << argv[1] << endl;
    return 1;
  }
  
  // Extract network info
  networkInfo * nwInfo = config->getNetworkInfo ();
  
  nasdsim_config::print (nwInfo);

  // UGLY: setting the global variable
  NASDSIM_SWITCH_TIME = nwInfo->switchTime;
  
  // Extract nasd info
  list <nasdInfo *> *nasdList = config->getNasdList ();
  if (nasdList == NULL || nasdList->size() == 0) {
    cerr << "error: no nasds to create!" << endl;
    exit(1);
  }
  
  list <nasdInfo *>::iterator nListIter;
  
  for (nListIter = nasdList->begin (); 
       nListIter != nasdList->end (); 
       nListIter++) 
    {
      nasdInfo *nInfo = *nListIter;
      
      nasdsim_config::print (nInfo);
      component_nasd *nasd = new component_nasd(nInfo->id,
						nInfo->bandwidths,
						nInfo->seekTime, 
						nInfo->cacheSize,
						nInfo->blockSize,
						nwInfo->bandwidth,
						nwInfo->headerSize,
						nasds_perceived_bw
						);
      comp.add_component_nasd (nInfo->id, nasd);
    }
  
  // Extract client info
  list <clientInfo *> *clientList = config->getClientList ();
  ASSERT (clientList != NULL);
  ASSERT (clientList->size () > 0);
  
  // Construct maxCreditHash indexed on nasdId
  hash_map <sim_component_id_t, int> maxCreditHash;
  
  for (nListIter = nasdList->begin (); 
       nListIter != nasdList->end (); 
       nListIter++) 
    {
      nasdInfo *nInfo = *nListIter;
      int n = nInfo->queueSize / clientList->size ();
      maxCreditHash [nInfo->id] = n;
    }
  
  // Create each client
  list <clientInfo *>::iterator cListIter;
  
  for (cListIter = clientList->begin (); 
       cListIter != clientList->end (); 
       cListIter++) 
    {
      clientInfo *cInfo = *cListIter;
      nasdsim_config::print (cInfo);
      component_client *client = new component_client (cInfo->id,
						       maxCreditHash,
						       cInfo->filename,
						       nwInfo->bandwidth,
						       nwInfo->headerSize,
						       true,
						       true);
      comp.add_component_client(cInfo->id, client);
      
    }
  
  // dispose off config object
  delete config;
  
  // ----------------------------------------------------------------------------------------
  run_simulation(comp, nasds_perceived_bw, argv[1]);
  
  return 0;
}


void
run_simulation(
	       components &comp,
	       perceived_map_t &perceived_bw,
	       const char *config_filename
	       )	       
{
  event_heap_t event_heap;          // a heap to hold simulation events.
  simtime_t    current_time;        // current simulation time
  event *      current_event;       // event to be simulated
  events_t *   next_events;         // events to be added to the heap from a simulation
                                    // of an event
  int          total_requests = 0;  // # of requests from the tracefile
  simtime_t    dump_interval  = 1;  // interval to dump info
  simtime_t    next_interval_dump_time = 0;
  
  simtime_t total_request_time = 0, last_perceived_computation = 0;

#if defined(NASD_REMOTE_CACHING) && defined(DUMP_GC_STATS)
  ofstream *cache_file = open_cache_stats(config_filename, comp);
  if (cache_file == NULL || !(*cache_file)) {
    cerr << "couldn't open cache file..." << endl;
    return;
  }
#endif

  nasds_list_t nasds = comp.getNasdList();
  perceived_map_t::iterator pbw_iter;

  // starting time of the simulator ought to be 0
  current_time = 0;

  // prime the event heap (new version)
  const clients_list_t clients = comp.getClientList ();
  clients_list_iter_t clients_iter;
  
  // prime the event heap (new version) -----------------------------------------------
  
  bool moreEvents = true;
  
  for (simtime_t tmpTime = 0; total_requests == 0 && moreEvents; tmpTime++) 
    {
      moreEvents = false;
      
      for (clients_iter = clients.begin(); clients_iter != clients.end();
	   clients_iter++)
	{
	  total_requests += add_events(event_heap, *(clients_iter->second), 
				       tmpTime, current_time); /**/
	  
	  if (!moreEvents) {
	    moreEvents = (clients_iter->second)->hasMoreEvents();
	  }
	}
    }
  //---------------------------- end prime --------------------------------------------
  
  // main simulation loop
  //
  // steps taken by this loop:
  //  1. get an event
  //  2. simulate the event
  //  3. check for a terminal event generated by the simulation
  //  4. get more events
  //
  while (!event_heap.empty())
   {
     // get the event to process
     current_event = event_heap.top();
     event_heap.pop();
     
     // Advancing the real time of the simulator
     current_time = current_event->get_start_time();
     
     // dump out interval info
     if (current_time >= next_interval_dump_time) {
       for (clients_iter = clients.begin(); clients_iter != clients.end();
	    clients_iter++)
	 clients_iter->second->dumpIntervalInfo (current_time);
     
       // obtain next time to dump
       while (current_time >= next_interval_dump_time) 
	 next_interval_dump_time += dump_interval;
     }
     
#if defined(DEBUG)
     cout << "simulating: ";
     current_event->print(cout);
     cout << endl;
#endif
     
     // simulate the event
     next_events = current_event->simulate(comp);
     
     // see if current_event generated a terminal event
     if (!next_events->empty() &&
	 dynamic_cast<event_empty *>(next_events->front()) != NULL) 
       {
	 // Advancing the real time of the simulator
	 // (04/30/2000) The only reason the real time is advanced here is to calculate
	 // total_request_time below. There is no need to advance time here as 
	 // total_request_time can be calculated by obtaining the start time of this
	 // empty event. However, we leave the code here.
	 current_time = next_events->front()->get_start_time();
	 
#if defined(DEBUG)
	 cout << "total request time: "
	      << current_time -
	   next_events->front()->get_injection_time() << endl;
#endif
	 total_request_time +=
	   (current_time - next_events->front()->get_injection_time());
	 
	 delete next_events->front();
	 delete next_events;
	 
	 // see if more events can be added to the event queue
	  if (event_heap.empty()) {
	    
	    // -----------------------------------------------------------------------
	    bool mEvents = false;
	    
	    for (clients_iter = clients.begin();
		 !mEvents && clients_iter != clients.end(); clients_iter++)
	      {
		mEvents = (clients_iter->second)->hasMoreEvents ();
	      }
	    
	    if (mEvents) {
	      
	      int num_events = 0;
	      
	      for (simtime_t tmpTime = current_time + 0.5; 
		   num_events == 0;
		   tmpTime += 5) 
		{
		  for (clients_iter = clients.begin();
		       clients_iter != clients.end(); clients_iter++)
		    {
		      num_events += add_events(event_heap,
					       *(clients_iter->second),
					       tmpTime, current_time); /**/
		    }
		}
	      total_requests += num_events;
	    }
	    // -----------------------------------------------------------------------
	    
	  }
       }
     else 
       {
	 // add the events from the next_events queue to the heap
	 while (!next_events->empty()) {
	   event_heap.push(next_events->front());
	   next_events->pop();
	 }
	 delete next_events;
       }
     
     delete current_event;
     
     // Get events from everybody
     for (clients_iter = clients.begin(); clients_iter != clients.end();
	  clients_iter++)
       {   
	 total_requests += add_events(event_heap, *(clients_iter->second),
				      event_heap.top()->get_start_time(),
				      current_time
				      );
       }
     
     // update perceived bandwidth of each NASD, if needed
     if (last_perceived_computation + 0.1 < current_time) {
       for (nasds_list_iter_t nasds_iter = nasds.begin();
	    nasds_iter != nasds.end(); nasds_iter++) {
	 nasds_iter->second->compute_perceived_bw();
       }
#if defined(NASD_REMOTE_CACHING) && defined(DUMP_GC_STATS)
       dump_cache_stats(cache_file, comp, current_time);
#endif

       last_perceived_computation = current_time;
     }
     
#if defined(DEBUG)
     cout << "Perceived bw's" << endl;
     for (pbw_iter = perceived_bw.begin(); pbw_iter != perceived_bw.end();
	  pbw_iter++) {
       cout << pbw_iter->first << "  " << pbw_iter->second.perceived_bw
	    << endl;
     }
     cout << endl;
#endif

#if defined(NASD_REMOTE_CACHING) && defined(DUMP_GC_STATS)
     dump_cache_stats(cache_file, comp, current_time);
#endif
   }
  
  print_stats(comp, current_time, total_requests, total_request_time);
}

void
print_stats(
	    const components &comp,
	    simtime_t current_time,
	    int total_requests,
	    simtime_t total_request_time
	    )
{
  // display final statistics
  cout << "NASDsim results" << endl;
  cout << "---------------" << endl;
  cout << endl;
  cout << "Components stats:" << endl;
  comp.dump_stats(cout);
  cout << endl;
  cout << "Total trace requests: " <<  total_requests << endl;
  cout << "Total time: " << current_time << endl;
  cout << "Average request service time: "
       << total_request_time/(double)total_requests << endl;  
}


/**
 * event_heap   - main event heap of the simulator
 * client       - client object
 * tmp_time     - time looking ahead to get more events
 * current_time - current time of the simulator 
 * returns      - none 
 */
int
add_events(
	   event_heap_t &event_heap,
	   component_client &client,
	   simtime_t tmp_time,
	   simtime_t current_time         /**/
	   )
{
  int total_requests = 0;

  // From each client obtain events that need to be scheduled at this time
  // For now there is just one client
  
  list <event *> * eventList = client.getEventList (tmp_time, current_time); /**/
  
  list <event *>::iterator eListIter;
  
  for (eListIter = eventList->begin (); eListIter != eventList->end (); eListIter++) 
    {
      ASSERT((*eListIter)->get_start_time() >= (*eListIter)->get_injection_time());
      
      // Modify ---------------------------------
      // (04/30/2000) removed
      // This was a quick fix to set the injection time correctly. Now we do it in 
      // the client.
      //
      //(*eListIter)->set_start_time (current_time);
      //(*eListIter)->set_injection_time (current_time);
      //-----------------------------------------
      
      event_heap.push (const_cast <event *> (*eListIter));
      total_requests++;
    }
  
  // Get rid of eventList
  delete eventList;

  return total_requests;
}

#if defined(NASD_REMOTE_CACHING) && defined(DUMP_GC_STATS)
ofstream *
open_cache_stats(
		 const char *config_filename,
		 components &comp
		 )
{
  nasds_list_t nasds = comp.getNasdList();
  nasds_list_iter_t nasds_iter;
  char *filename
    = new char[strlen(config_filename) + strlen("cache_file-") + 1];
  ofstream *cache_file;

  strcpy(filename, "cache_file-");
  strcat(filename, config_filename);

  cache_file = new ofstream(filename);

  if (cache_file != NULL && !cache_file->fail()) {
    (*cache_file) << "Time ";
    for (nasds_iter = nasds.begin();
	 nasds_iter != nasds.end();
	 nasds_iter++) {
      (*cache_file) << "nasd " << nasds_iter->first << " ";
    }
    (*cache_file) << endl;
  }

  return cache_file;
}

void
dump_cache_stats(
		 ofstream *output,
		 components &comp,
		 simtime_t current_time
		 )
{
  nasds_list_t nasds = comp.getNasdList();
  nasds_list_iter_t nasds_iter1, nasds_iter2;
  int total_bytes;

  (*output) << current_time << " ";
  for (nasds_iter1 = nasds.begin();
       nasds_iter1 != nasds.end();
       nasds_iter1++) {
    total_bytes = 0;
    for (nasds_iter2 = nasds.begin();
	 nasds_iter2 != nasds.end();
	 nasds_iter2++) {
      total_bytes +=
	nasds_iter2->second->get_cache_global_num_entries(nasds_iter1->first);
    }
    (*output) << total_bytes << " ";
  }
  (*output) << endl;
}
#endif
