//
// Description:
//  The vast majority (and I think all) NASD operations require some kind of
//  parameters. Rather than pass the parameters as basic C++ types, which
//  is cumbersome and less than easy to extend, when a new event in the
//  request event chain is created, param objects are passed. This is much
//  easier to handle as the simulator changes to meet new needs.
//

// $Id: nasdsim_param.h,v 1.12 2000/05/09 01:05:56 bforney Exp $


#if !defined(NASDSIM_PARAM_H)
#define NASDSIM_PARAM_H

#include <iostream>
#include "nasdsim_types.h"
#include <list>

class param {
 public:
  param(sim_component_id_t cId, sim_component_id_t nId) 
    : client_id (cId), nasd_id (nId) {}
  ~param() {}
  
  // accessors
  sim_component_id_t get_client_id (void) const { return client_id; }
  sim_component_id_t get_nasd_id (void) const { return nasd_id; }

  
 protected:
  sim_component_id_t client_id;
  sim_component_id_t nasd_id;

};

class param_read : public param {
 public:
  param_read(sim_object_t object, sim_offset_t offset, sim_len_t length, 
	     sim_component_id_t cId, sim_stream_id_t strId, sim_component_id_t nId,
	     sim_request_id_t request_id, sim_len_t total_length,
	     bool partial = false) :
    param (cId, nId), stream_id (strId), object_(object), offset_(offset), 
    length_(length), request_id_(request_id), total_length_ (total_length), 
    partial_(partial), returned_(false) {};
  ~param_read() {};
  
  sim_object_t get_object(void) const { return object_; }
  sim_offset_t get_offset(void) const { return offset_; }
  sim_len_t get_length(void) const { return length_; }
  sim_len_t get_total_length (void) const { return total_length_; }
  sim_stream_id_t get_stream_id (void) const { return stream_id; }  
  sim_request_id_t get_request_id (void) const { return request_id_; }
  bool get_partial (void) const { return partial_; }
  bool get_returned(void) const { return returned_; }
  
  void set_length (sim_len_t length) { length_ = length; }
  void inc_length (sim_len_t l) { length_ += l; }
  void set_partial (bool partial) { partial_ = partial; }
  void set_returned(bool returned) { returned_ = returned; }
  
  void print(ostream &output) const;
  
 private:
  
  sim_stream_id_t stream_id;
  sim_object_t object_;
  sim_offset_t offset_;
  sim_len_t length_;
  sim_request_id_t request_id_;
  sim_len_t total_length_;                             // KLUDGE: Used while returning  
  bool partial_;                                       // Yet another KLUDGE
  bool returned_;
};

class param_write : public param {
 public:
  param_write(sim_object_t object, sim_offset_t offset, sim_len_t length,
	      sim_component_id_t cId, sim_stream_id_t strId, sim_component_id_t nId) :
    param (cId, nId), stream_id (strId), object_(object), offset_(offset),
    length_(length) {}
  ~param_write() {};

  sim_object_t get_object(void) const { return object_; }
  sim_offset_t get_offset(void) const { return offset_; }
  sim_len_t get_length(void) const { return length_; }
  sim_stream_id_t get_stream_id (void) const { return stream_id; }
    
  void print(ostream &output) const;

 private:
  sim_stream_id_t stream_id;
  sim_object_t object_;
  sim_offset_t offset_;
  sim_len_t length_;
};

// This paramter object accompanies all events that occur during communication between
// nasds (event_N2N...). It just specifies a message being transferred between a source
// nasd to a destination nasd.
//
// The paramter object used in events will actually be a sub-class of this object. The
// type of the sub-class will be deduced at the destination by doing the appropriate
// cast. Based on the type of sub-class the correct action will be taken.
class param_nasd_2_nasd {
 public:
    
  param_nasd_2_nasd (sim_component_id_t source_nasd,
		     sim_component_id_t  destination_nasd,
		     sim_len_t           length) :
    source_nasd_(source_nasd), destination_nasd_(destination_nasd), 
    length_(length) {};
  
  virtual ~param_nasd_2_nasd () {};
  
  sim_len_t get_length (void) const { return length_; }
  sim_component_id_t get_source_nasd (void) { return source_nasd_; }
  sim_component_id_t get_destination_nasd (void) { return destination_nasd_; }
  
  virtual void print (ostream &output) const = 0;
  
 protected:
  
  sim_component_id_t source_nasd_;
  sim_component_id_t destination_nasd_;
  sim_len_t          length_;
};

// Scenario: A parent nasd is evicting some blocks and requesting a remote nasd to
// store them. 
// o It will use param_parent_eviction_request type parameter to send the request.
// o A list of blocks to be evicted will be sent.
// o After sending the parent nasd will suspend and do nothing until it receives an
//   acknowledgement from the remote nasd.
// o Parameter type used in the response is param_parent_eviction_reply.

// CAUTION: The class does not destroy 
// o block_list_
class param_parent_eviction_request : public param_nasd_2_nasd 
{
 public:
    
    param_parent_eviction_request (sim_component_id_t   source_nasd,
				   sim_component_id_t   destination_nasd,
				   sim_len_t            length,
				   block_list_t *block_list) :
	param_nasd_2_nasd (source_nasd, destination_nasd, length),
	block_list_ (block_list) {};
    ~param_parent_eviction_request () {};
    
    void print (ostream &output) const;
    
    block_list_t *get_block_list (void) const { return block_list_; }
    
    void set_block_list (block_list_t *l) { block_list_ = l; }

 private:
    
    block_list_t *block_list_;
};

class param_parent_eviction_reply : public param_nasd_2_nasd
{
 public:
    
    param_parent_eviction_reply (sim_component_id_t source_nasd,
				 sim_component_id_t destination_nasd,
				 sim_len_t          length,
				 bool               reply) :
	param_nasd_2_nasd (source_nasd, destination_nasd, length),
	reply_ (reply) {};
    ~param_parent_eviction_reply () {};
    
    void print (ostream &output) const;
    
    bool get_reply (void) const { return reply_; }

 private:
    
    bool reply_;
};

// Scenario: parent nasd is forwarding a partial read request to a remote nasd that 
// contains some blocks in its cache.
// o A copy of param_read object is made
// o Its length is adjusted
// o Its partial flag is set to true
// o The part of the request that the parent can satisfy it goes ahead and sends back
//   to the client
// o The other part is sent to remote nasd using param_read_forward_request
// o Remote nasd is expected to honor the request and is responsible of sending the
//   response to the client.
    
// CAUTION: The class does not destroy 
// o parameter_
// o block_list_ (if any)
class param_read_forward_request : public param_nasd_2_nasd 
{
 public:
    
    param_read_forward_request (sim_component_id_t   source_nasd,
				sim_component_id_t   destination_nasd,
				sim_len_t            length,
				param_read *parameter,
				block_list_t *block_list = NULL) :
	param_nasd_2_nasd (source_nasd, destination_nasd, length),
	parameter_(parameter), block_list_ (block_list) {};
    ~param_read_forward_request () {};
    
    void print (ostream &output) const;
    
    block_list_t *get_block_list (void) const { return block_list_; }
    param_read * get_param_read (void) const { return parameter_; }
    
    void set_block_list (block_list_t *l) { block_list_ = l; }    
    void set_param_read (param_read *parameter) { parameter_ = parameter; }

 private:
    
    param_read *parameter_;
    block_list_t *block_list_;
};

#if 0
// Scenario: Remote nasd decides to throw away some blocks that it has cached for a 
// parent nasd.
// o It uses param_remote_eviction_request object to convey the information to the 
//   parent nasd.
// o The remote nasd suspends all activity until it gets an acknowledgement from the
//   parent nasd. It is only after this that it throws away the victim blocks. However
//   while waiting for the response from the parent nasd the remote nasd is expected
//   to honor requests that come for blocks that are available in its cache. In other
//   words any request that does not cause a change in the state of the cache is 
//   honored. The reply from the parent nasd will come in the form of
//   param_remote_eviction_reply

// CAUTION: The class does not destroy 
// o block_list_
class param_remote_eviction_request : public param_nasd_2_nasd 
{
 public:
    
    param_remote_eviction_request (sim_component_id_t   source_nasd,
				   sim_component_id_t   destination_nasd,
				   sim_len_t            length,
				   block_list_t *block_list) :
	param_nasd_2_nasd (source_nasd, destination_nasd, length),
	block_list_ (block_list) {};
    ~param_remote_eviction_request () {};
    
    void print (ostream &output) const;
    
    block_list_t *get_block_list (void) const { return block_list_; }
    
    void set_block_list (block_list_t *l) { block_list_ = l; }    

 private:
    
    block_list_t *block_list_;
};

class param_remote_eviction_reply : public param_nasd_2_nasd
{
 public:
    
    param_remote_eviction_reply (sim_component_id_t source_nasd,
				 sim_component_id_t destination_nasd,
				 sim_len_t          length,
				 bool               reply) :
      param_nasd_2_nasd (source_nasd, destination_nasd, length),
      reply_ (reply) {};
    ~param_remote_eviction_reply () {};
    
    void print (ostream &output) const;
    
    bool get_reply (void) const { return reply_; }
    
 private:
    
    bool reply_;
};
#endif

// Scenario: Parent nasd wants to inform a remote nasd to invalidate the cached blocks
// that it has on behalf the parent nasd. 
// o param_parent_invalidate_request is used to convey this information.

// CAUTION: The class does not destroy 
// o block_list_
class param_parent_invalidate_request : public param_nasd_2_nasd 
{
 public:
    
    param_parent_invalidate_request (sim_component_id_t   source_nasd,
				     sim_component_id_t   destination_nasd,
				     sim_len_t            length,
				     block_list_t *block_list) :
	param_nasd_2_nasd (source_nasd, destination_nasd, length),
	block_list_ (block_list) {};
    ~param_parent_invalidate_request () {};
    
    void print (ostream &output) const;
    
    block_list_t *get_block_list (void) const { return block_list_; }
    
    void set_block_list (block_list_t *l) { block_list_ = l; }    

 private:
    
    block_list_t *block_list_;
};

//
// This type is used when an owner/parent NASD forwards requests to NASDs
// that it believes will satisfy the request, but these target NASDs do
// not have 
//
class param_returned_read_request : public param_nasd_2_nasd
{
 public:
    
    param_returned_read_request (sim_component_id_t   source_nasd,
				 sim_component_id_t   destination_nasd,
				 sim_len_t            length,
				 block_list_t *block_list,
				 param_read *parameter) :
      param_nasd_2_nasd(source_nasd, destination_nasd, length),
      block_list_(block_list) { parameter_ = new param_read(*parameter); }
    ~param_returned_read_request () { delete parameter_; }
    
    void print (ostream &output) const;
    param_read * get_param_read (void) const { return parameter_; }

    block_list_t *get_block_list (void) const { return block_list_; }
    
    void set_block_list (block_list_t *l) { block_list_ = l; }    

 private:
    
    block_list_t *block_list_;
    param_read *parameter_;
};

#endif // !defined(NASDSIM_PARAM_H) 



